"""metrics_manager.py
============
Functionality related to managing metrics
"""

import requests
from collections.abc import Iterable
from .util import raise_resp_exception_error, ElementType
from typing import (
    Any,
    Union,
    List,
    Dict,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
)
from typing_extensions import TypedDict
from termcolor import colored
from base64 import b64encode
import json

if TYPE_CHECKING:
    from .client import Client

# TODO: Better typing
QueryEntry = List[Tuple[str, str]]
QueryEntries = List[QueryEntry]
QueryOrdering = str


class ConfusionsOptsRequired(TypedDict, total=True):
    confidence_threshold: Union[float, int]
    iou_threshold: Union[float, int]
    queries: QueryEntries
    ordering: QueryOrdering


class ConfusionsOpts(ConfusionsOptsRequired, total=False):
    cur_offset: int
    max_results: int
    include_background: bool


class ConfusionsResultDict(TypedDict):
    cur_offset: int
    num_rows: int
    rows: List[Dict[str, Any]]


T_ = TypeVar("T_")


def flatten(nested: List[List[T_]]) -> List[T_]:
    return [item for sublist in nested for item in sublist]


class MetricsManager:
    """A manager for interacting with metrics within a given project.

    Contains the following constants:

    .. code-block:: text

        MetricsManager.ORDER_CONF_DESC
        MetricsManager.ORDER_CONF_ASC
        MetricsManager.ORDER_IOU_DESC
        MetricsManager.ORDER_IOU_ASC
        MetricsManager.ORDER_BOX_SIZE_DESC
        MetricsManager.ORDER_BOX_SIZE_ASC

    Args:
        client (Client): An Aquarium Learning Python Client object.
        project_id (str): The project id associated with this manager.
    """

    ORDER_CONF_DESC = "inference_data.confidence__by__desc"
    ORDER_CONF_ASC = "inference_data.confidence__by__asc"
    ORDER_IOU_DESC = "iou__by__desc"
    ORDER_IOU_ASC = "iou__by__asc"
    ORDER_BOX_SIZE_DESC = "box_size__by__desc"
    ORDER_BOX_SIZE_ASC = "box_size__by__asc"

    def __init__(self, client: "Client", project_id: str) -> None:
        print(
            colored(
                "\nThe metrics manager is in an alpha state, and may experience breaking changes in future updates\n",
                "yellow",
            )
        )

        self.client = client
        self.project_id = project_id
        self.project_info = self.client.get_project(project_id)

        # Matches logic in /all_detection_metrics
        inference_class_map = [
            x
            for x in self.project_info["label_class_map"]
            if "train_name" not in x or x["train_name"] != ""
        ]
        inference_class_names = [
            x.get("train_name") or x.get("name") for x in inference_class_map
        ]

        # Pre-computed metrics are normalized to always be lowercased to be case insensitive
        inference_class_names = [x.lower() for x in inference_class_names]
        inference_class_set = set(inference_class_names)

        # Make sure it has background
        self.inference_classes_nobg = sorted(list(inference_class_set))
        self.inference_classes = self.inference_classes_nobg + ["background"]
        self.inference_classes_set = set(self.inference_classes)

    def make_cell_query(self, gt: str, inf: str) -> QueryEntry:
        """Make a query entry for a specific confusion of gt as as inf.

        Args:
            gt (str): The classname of the ground truth class.
            inf (str): The classname of the inference class.

        Returns:
            QueryEntry: Result query entry.
        """
        if (
            gt not in self.inference_classes_set
            or inf not in self.inference_classes_set
        ):
            raise Exception("Invalid classname provided.")

        return [(gt, inf)]

    def make_correct_query(self) -> QueryEntry:
        """Make a query entry for all correct detections/classifications (gt == inf)

        Returns:
            QueryEntry: Result query entry.
        """

        return [(name, name) for name in self.inference_classes_nobg]

    def make_confusions_query(self) -> QueryEntry:
        """Make a query entry for all confusions (gt != inf)

        This will only include cases where two matched detections exist, not
        for false positive or false negative detections.

        Returns:
            QueryEntry: Result query entry.
        """

        acc = []
        for gt_name in self.inference_classes_nobg:
            for inf_name in self.inference_classes_nobg:
                if gt_name == inf_name:
                    continue

                acc.append((gt_name, inf_name))

        return acc

    def make_false_positives_query(self) -> QueryEntry:
        """Make a query entry for all false positive detections.

        These are cases without corresponding ground truth detections
        for an inference detection.

        Returns:
            QueryEntry: Result query entry.
        """

        return [("background", name) for name in self.inference_classes_nobg]

    def make_false_negatives_query(self) -> QueryEntry:
        """Make a query entry for all false negative detections.

        These are cases without corresponding inference detections
        for a ground truth detection.

        Returns:
            QueryEntry: Result query entry.
        """

        return [(name, "background") for name in self.inference_classes_nobg]

    def make_confused_as_query(self, name: str) -> QueryEntry:
        """Make a query entry for all confusions as name. (inf == name, gt != inf)

        This will only include cases where two matched detections exist, not
        for false positive or false negative detections.

        Returns:
            QueryEntry: Result query entry.
        """

        if name not in self.inference_classes_set:
            raise Exception("Invalid classname provided.")

        acc = []
        for other_name in self.inference_classes_nobg:
            if other_name == name:
                continue
            acc.append((other_name, name))

        return acc

    def make_confused_from_query(self, name: str) -> QueryEntry:
        """Make a query entry for all confusions from name. (gt == name, gt != inf)

        This will only include cases where two matched detections exist, not
        for false positive or false negative detections.

        Returns:
            QueryEntry: Result query entry.
        """

        if name not in self.inference_classes_set:
            raise Exception("Invalid classname provided.")

        acc = []
        for other_name in self.inference_classes_nobg:
            if other_name == name:
                continue
            acc.append((name, other_name))

        return acc

    def _encode_query(self, query: QueryEntry) -> str:
        """This function takes in a full-form Query Entry (list of pairs), and returns a short string representation.

        To do so, it replaces explicit lists of inclusions with lists of exclusions.
        For example, assume you have 100 classes and want every time class X is confused as something else.
        This is equivalent to a row of the confusion matrix, minus the cell on the diagonal.

        If we list all included cells, that's 99 unique pairs. Instead, we can represent membership
        as "this row, except where cell[0] == X and cell[1] == X".

        This function takes the input and replaces cells with more efficient row/column-except representations.

        Then, it takes that shorter list and jsonifies -> b64 encodes so it can be easily transmitted and stored.

        Args:
            query (QueryEntry): Full-form query to be compressed / encoded

        Returns:
            str: base64 encoded form of the json string describing the query.
        """
        REQ_RATIO = 3

        def pairKey(a: str, b: str) -> str:
            return f"{a}__aq__{b}"

        def splitKey(s: str) -> List[str]:
            return s.split("__aq__")

        unassigned = set([pairKey(x[0], x[1]) for x in query])
        result = []

        for inf_name in self.inference_classes:
            yes = []
            no = []
            for gt_name in self.inference_classes:
                key = pairKey(gt_name, inf_name)
                if key in unassigned:
                    yes.append(gt_name)
                else:
                    no.append(gt_name)

            if len(yes) > (len(no) * REQ_RATIO):
                res = ["inf_except_gt", inf_name]
                res.extend(no)
                for gt_name in self.inference_classes:
                    to_remove_key = pairKey(gt_name, inf_name)
                    if to_remove_key in unassigned:
                        unassigned.remove(to_remove_key)
                result.append(res)

        for gt_name in self.inference_classes:
            yes = []
            no = []
            for inf_name in self.inference_classes:
                key = pairKey(gt_name, inf_name)
                if key in unassigned:
                    yes.append(gt_name)
                else:
                    no.append(gt_name)

            if len(yes) > (len(no) * REQ_RATIO):
                res = ["gt_except_inf", gt_name]
                res.extend(no)
                for inf_name in self.inference_classes:
                    to_remove_key = pairKey(gt_name, inf_name)
                    if to_remove_key in unassigned:
                        unassigned.remove(to_remove_key)
                result.append(res)

        for remaining in unassigned:
            gt_name, inf_name = splitKey(remaining)
            result.append(["pair", gt_name, inf_name])

        stringified = json.dumps(result)
        encoded = b64encode(stringified.encode("utf-8")).decode("utf-8")
        return encoded

    def fetch_confusions(
        self, dataset_id: str, inferences_id: str, opts: ConfusionsOpts
    ) -> ConfusionsResultDict:
        """Fetch confusions for a given dataset + inference set pair.

        The options define the parameters of the query. The options are a dictionary of the form:

        .. code-block:: text

            {
                'confidence_threshold': Union[float, int],
                'iou_threshold': Union[float, int],
                'queries': List[QueryEntry],
                'ordering': QueryOrdering,
                'max_results': int (optional, defaults to 10,000),
                'cur_offset': int (optional, defaults to 0),
            }


        Confidence and iou thresholds can be any multiple of 0.1 between 0.0 and 1.0.
        Queries are a list of queries produced from helper methods such as

            metrics_manager.make_confusions_query()

        Ordering is defined by constants on the class object, such as:

            metrics_manager.ORDER_CONF_DESC

        Args:
            dataset_id (str): The dataset id.
            inferences_id (str): The inference set id.
            opts (ConfusionsOpts): Options for the query.

        Returns:
            ConfusionsResultDict: All of your query results, up to max_results (default = 10k)
        """

        datasets = self.client.get_datasets(self.project_id, include_archived=False)
        existing_dataset_ids = [dataset.get("id") for dataset in datasets]

        inferences_dataset_id = inferences_id
        if not inferences_dataset_id.startswith("inferences_"):
            inferences_dataset_id = "_".join(["inferences", dataset_id, inferences_id])

        if (
            inferences_dataset_id not in existing_dataset_ids
            or dataset_id not in existing_dataset_ids
        ):
            raise Exception("Invalid IDs provided.")

        base_dataset_info = next(x for x in datasets if x.get("id") == dataset_id)
        inferences_info = next(
            x for x in datasets if x.get("id") == inferences_dataset_id
        )

        is_streaming = base_dataset_info.get("is_streaming", False)
        latest_window = str((inferences_info.get("ingest_agg_windows") or [0])[-1])
        bbox_metrics_metadata = inferences_info.get("bbox_metrics_metadata")
        if bbox_metrics_metadata is None:
            raise Exception("No valid metrics computed.")

        ######################################################################
        # Load (potentially windowed / versioned) data table
        ######################################################################
        dataset_table = ".".join([self.project_id, dataset_id])
        inference_set_table = ".".join([self.project_id, inferences_dataset_id])

        # TODO: Support streaming datasets
        table_query_params: Dict[str, Union[bool, str]] = {
            "dataset": dataset_table,
            "inference_set": inference_set_table,
            "metrics_table": bbox_metrics_metadata.get("metrics_table"),
            "is_streaming": is_streaming,
            "project": self.project_id,
            "window": latest_window,
            "query": "{}",
        }

        url = "/joined_filtered_table"
        r = requests.get(
            self.client.api_endpoint + url,
            headers=self.client._get_creds_headers(),
            params=table_query_params,
        )

        raise_resp_exception_error(r)
        result = r.json()

        labels_table_filtered = result["dest_table"]

        ######################################################################
        # Load confusion matrix
        ######################################################################

        for field in ["confidence_threshold", "iou_threshold", "queries", "ordering"]:
            if field not in opts:
                raise Exception(f"Required field {field} not in opts.")

        flattened_query_cells = flatten(opts["queries"])
        encoded_query = self._encode_query(flattened_query_cells)

        # TODO: Make labels_table_filtered the results of calling `/api/v1/query`

        default_payload_args = {
            "active_metric_type": "default",
            "associations_table": bbox_metrics_metadata["metrics_table"],
            "cur_offset": 0,
            "include_background": False,
            "inferences_table": inference_set_table,
            "labels_table": dataset_table,
            "labels_table_filtered": labels_table_filtered,
            "max_results": 10000,
            "other_associations_table": None,
        }

        payload = {**default_payload_args, "encoded_filter_classes": encoded_query}
        for k, v in opts.items():
            if k == "queries":
                continue

            payload[k] = v  # type: ignore

        if int(payload["iou_threshold"] * 100) % 10 != 0:
            raise Exception("iou_threshold must be a multiple of 0.1")
        if int(payload["confidence_threshold"] * 100) % 10 != 0:
            raise Exception("confidence_threshold must be a multiple of 0.1")

        url = "/confusion_matrix_query"
        r = requests.post(
            self.client.api_endpoint + url,
            headers=self.client._get_creds_headers(),
            json=payload,
        )
        raise_resp_exception_error(r)

        result = r.json()
        if "dest_table" in result:
            del result["dest_table"]

        return result



"""collection_client.py
============
The extended client module for collection campaigns.
"""

import os
import datetime
from random import random
import sys
import time
import uuid
import math
import json
import pandas as pd
from requests.packages.urllib3.util import retry
import pyarrow as pa
from uuid import uuid4
from io import IOBase
from tempfile import NamedTemporaryFile
from tqdm import tqdm
from google.resumable_media.requests import ChunkedDownload
from google.resumable_media.common import InvalidResponse, DataCorruption
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    List,
    Dict,
    Tuple,
    IO,
    BinaryIO,
    TypeVar,
    Iterator,
)
from typing_extensions import Literal
from queue import PriorityQueue

from requests.adapters import prepend_scheme_if_needed

from .util import (
    requests_retry,
    raise_resp_exception_error,
    _is_one_gb_available,
    _upload_local_files,
    create_temp_directory,
    mark_temp_directory_complete,
    MAX_CHUNK_SIZE,
    MAX_FRAMES_PER_BATCH,
    DEFAULT_SAMPLING_THRESHOLD,
)
from .client import Client
from .dataset import LabeledDataset, LabeledFrame
from .inference import Inferences, InferencesFrame
from .sampling_agent import RandomSamplingAgent, SamplingScoreDict, SamplingAgent

from .embedding_distance_sampling import EmbeddingDistanceSamplingAgent

CollectionAssetType = Literal["sample_frame", "embedding"]
SortStatus = Literal["UNSORTED", "DISCARDED", "ACCEPTED"]
SimilarityResultProcessingStatus = Literal["ENQUEUED", "RUNNING", "DONE"]

DEFAULT_STATISTICS = {
    "n": 0,
    "min": float("inf"),
    "max": float("-inf"),
    "mean": 0,
    "variance": 0,
    "m2": 0,
}


class CollectionClient(Client):
    """Client class that interacts with the Aquarium REST API.
    Also handles extra work around collecting samples for collection campigns

    Args:
        api_endpoint (str, optional): The API endpoint to hit. Defaults to "https://illume.aquariumlearning.com/api/v1".
    """

    active_coll_camp_summaries: Dict[str, Any]  # Dictionary of campaign ID to summary
    camp_id_to_sample_agent: Dict[str, SamplingAgent]
    # TODO: This Any should be Optional[str], but the code doesn't currently handle it
    frame_batch_uuid_to_temp_file_path: Dict[str, Any]

    # TODO: I think it should be SamplingScoreDict?
    # frame_batch_uuid_to_camp_id_to_probability_score_info: Dict[str, Dict[str, List[SamplingScoreDict]]]
    frame_batch_uuid_to_camp_id_to_probability_score_info: Dict[
        str, Dict[str, List[Any]]
    ]
    camp_id_to_gcs_frame_urls: Dict[str, List[str]]
    camp_id_to_gcs_embedding_urls: Dict[str, List[str]]
    sampled_frame_statistics: Dict[str, Any]
    most_similar_frame: Optional[Dict[str, Any]]
    most_similar_frame_campaign_id: Optional[str]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.active_coll_camp_summaries = {}

        self.sampling_agent = EmbeddingDistanceSamplingAgent
        self.camp_id_to_sample_agent = {}

        self.frame_batch_uuid_to_temp_file_path = {}
        self.frame_batch_uuid_to_camp_id_to_probability_score_info = {}

        # A mapping of campaign ID to the list of GCS URLs (one URL per batch of frames)
        self.camp_id_to_gcs_frame_urls = {}
        self.camp_id_to_gcs_embedding_urls = {}

        self.temp_file_path = create_temp_directory()

        self.num_frames_processed = 0
        self.sampled_frame_statistics = DEFAULT_STATISTICS
        self.most_similar_frame = None
        self.most_similar_frame_campaign_id = None

    def _clear_cached_sample_state(self) -> None:
        self.frame_batch_uuid_to_temp_file_path = {}
        self.frame_batch_uuid_to_camp_id_to_probability_score_info = {}
        self.camp_id_to_gcs_frame_urls = {}
        self.num_frames_processed = 0
        self.camp_id_to_gcs_embedding_urls = {}
        self.sampled_frame_statistics = DEFAULT_STATISTICS
        self.most_similar_frame = None
        self.most_similar_frame_campaign_id = None

    def _campaign_str(self, campaign: Dict[str, Any]) -> str:
        return f"Collection Campaign (Issue {campaign.get('issue_uuid')}, version {campaign.get('issue_version')})"

    def _save_content_to_temp(
        self,
        file_name_prefix: str,
        writefunc: Callable[[IO[Any]], bool],
        mode: str = "w",
    ) -> Optional[str]:
        """saves whatever the write function wants to a temp file and returns the file path

        Args:
            file_name_prefix (str): prefix for the filename being saved
            writefunc ([filelike): function used to write data to the file opened

        Returns:
            str or None: path of file or none if nothing written
        """

        if not _is_one_gb_available():
            raise OSError(
                "Attempting to flush dataset to disk with less than 1 GB of available disk space. Exiting..."
            )

        data_rows_content = NamedTemporaryFile(
            mode=mode, delete=False, prefix=file_name_prefix, dir=self.temp_file_path
        )
        data_rows_content_path = data_rows_content.name
        success = writefunc(data_rows_content)

        data_rows_content.seek(0)
        data_rows_content.close()

        return data_rows_content_path if success else None

    def _write_frames_to_file(
        self, frames: List[Dict[str, Any]], filelike: IO[Any]
    ) -> bool:
        """Write the frame content to a text filelike object (File handle, StringIO, etc.)

        Args:
            filelike (filelike): The destination file-like to write to.
        Returns:
            bool indicating whether write was successful
        """
        for frame in frames:
            filelike.write(json.dumps(frame) + "\n")

        # Check if something was written
        return filelike.tell() != 0

    def _write_embeddings_to_file(
        self, embeddings: List[Dict[str, Any]], filelike: IO[Any]
    ) -> bool:
        """Write the embeddings to a text filelike object (File handle, StringIO, etc.)

        Args:
            embeddings (List[Dict[str, Any]]): Keys are "embedding" and "task_id"
            filelike (filelike): The destination file-like to write to.
        Returns:
            bool indicating whether write was successful
        """
        embeddings_as_dict = {
            "task_id": pd.Series([x["task_id"] for x in embeddings]),
            "embedding": pd.Series([x["embedding"] for x in embeddings]),
        }
        df = pd.DataFrame(embeddings_as_dict)
        indexed_df = df.set_index("task_id")
        if indexed_df is None:
            return False

        table = pa.Table.from_pandas(indexed_df)

        ipc_writer = pa.ipc.new_file(
            filelike.name, table.schema, use_legacy_format=False  # type: ignore
        )
        ipc_writer.write_table(table)  # type: ignore
        ipc_writer.close()

        return True

    def download_to_file(self, signed_url: str, filelike: IO[Any]) -> bool:
        xml_api_headers = {
            "content-type": "application/octet-stream",
        }
        download = ChunkedDownload(signed_url, MAX_CHUNK_SIZE, filelike)
        while not download.finished:
            try:
                download.consume_next_chunk(requests_retry)
            except (InvalidResponse, DataCorruption, ConnectionError):
                if download.invalid:
                    continue
                continue
        return True

    def get_unlabeled_results_for_issue(
        self,
        *,
        issue_uuid: str,
        unlabeled_dataset_name: Optional[str],
        sort_status: Optional[SortStatus] = "UNSORTED",
        unlabeled_only: Optional[bool] = True,
        retrigger_search: Optional[bool] = False,
    ) -> List[Dict[str, Any]]:
        """returns similar results from unlabeled dataset search

        Args:
            issue_uuid (str): the uuid of the seed issue
            unlabeled_dataset_name (str): the name of the unlabeled search dataset
            sort_status (Optional[SortStatus]): filter the returned collection frames by status (default UNSORTED)
            unlabeled_only (Optional[bool]): return only collection frames that haven't been sent to labeling
            retrigger_search (Optional[bool]): retrigger similarity search (NOTE: this only works for searches against unlabeled datasets. It will poll + wait for the result)

        Returns:
            List[Dict[str, Any]]: List of CollectionFrames for a given issue
        """

        if not unlabeled_dataset_name and retrigger_search:
            raise Exception(
                "Can't retrigger search for Python client collection campaigns, only campaigns against unlabeled datasets"
            )

        collection_campaigns = self._get_all_campaigns(target_issue_uuids=[issue_uuid])
        if not collection_campaigns:
            raise Exception(
                f"Issue {issue_uuid} does not support similarity search because it does not have an associated collection campaign."
            )

        campaign_id = collection_campaigns[0]["id"]
        project_name = collection_campaigns[0]["project_name"]

        if retrigger_search:
            print(f"Retriggering unlabeled similarity search for issue {issue_uuid}...")

            search_payload = {
                "search_dataset_name": unlabeled_dataset_name,
                "collection_campaign_id": campaign_id,
            }

            r = requests_retry.post(
                self.api_endpoint
                + f"/projects/{project_name}/issues/{issue_uuid}/find_similar_elements",
                headers=self._get_creds_headers(),
                json=search_payload,
            )
            raise_resp_exception_error(r)

            # TODO: possibly insert sleep in case status isn't set right away?

            retry_count = 0
            processing_status: SimilarityResultProcessingStatus = "ENQUEUED"
            while processing_status != "DONE":
                if retry_count != 0:
                    sleep_time = 2**retry_count
                    print(f"Still running. Sleeping for {sleep_time} seconds...")
                    time.sleep(sleep_time)

                print("Polling for search status...")
                polling_params = {"search_dataset_name": unlabeled_dataset_name}
                r = requests_retry.get(
                    self.api_endpoint
                    + f"/projects/{project_name}/issues/{issue_uuid}/similar_element_info",
                    headers=self._get_creds_headers(),
                    params=polling_params,
                )

                raise_resp_exception_error(r)
                poll_result: Dict[str, Any] = r.json()
                processing_status = poll_result["similar_elts_processing_status"]
                retry_count += 1

            print("Similarity search complete, fetching results...")

        fetch_params: Dict[str, Any] = {
            "unlabeled_only": unlabeled_only,
            "sort_status": sort_status,
            "python_client": True,
        }

        if unlabeled_dataset_name:
            fetch_params["search_dataset"] = unlabeled_dataset_name

        r = requests_retry.get(
            self.api_endpoint
            + f"/projects/{project_name}/collection_frames/for_issue/{issue_uuid}",
            headers=self._get_creds_headers(),
            params=fetch_params,
        )

        raise_resp_exception_error(r)
        result: List[Dict[str, Any]] = r.json()
        return result

    def _read_rows_from_disk(self, file_path: str) -> List[Dict[str, Any]]:
        """reads temp files from disk and loads the dicts in them into memory

        Args:
            file_path (str): file path to read from

        Returns:
            List[Dict[str, Any]]: List of LabeledFrames in a dict structure
        """
        with open(file_path, "r") as frame_file:
            return [json.loads(line.strip()) for line in frame_file.readlines()]

    def _get_all_campaigns(
        self, target_project_names: List[str] = [], target_issue_uuids: List[str] = []
    ) -> List[Dict[str, Any]]:
        """Gets all collection campaign summaries (filtered to the target_project_names if provided)

        Args:
            target_project_names (Optional[List[str]], optional): List of project names whose collection campaigns should be sampled for
            target_issue_uuids (Optional[List[str]], optional): List of issue uuids whose collection campaigns should be sampled for. This takes precedence over 'target_project_names' and is mutually exclusive.

        Returns:
            List[Dict[str, Any]]: List of dicts containing collection campaign summaries
        """
        params = {}
        if target_issue_uuids:
            params["target_issue_uuids"] = target_issue_uuids
        elif target_project_names:
            params["target_project_names"] = target_project_names

        r = requests_retry.get(
            self.api_endpoint + "/collection_campaigns/summaries",
            headers=self._get_creds_headers(),
            params=params,
        )

        raise_resp_exception_error(r)
        result: List[Dict[str, Any]] = r.json()
        return result

    def _upload_collection_assets_to_gcs(
        self,
        frame_batch_uuid: str,
        campaign_id: str,
        assets: Any,
        writefunc: Callable[[Any, IO[Any]], bool],
        asset_type: CollectionAssetType,
    ) -> List[str]:
        """takes frames for collection and posts it to the API in batches

        Args:
            frame_batch_uuid (str): The local frame batch these frames were sampled from
            campaign_id (str): The collection campaign id
            assets (Dict[str, Any]): Dict structure containing collection assets for a campaign to post
            asset_type (CollectionAssetType): The type of asset being uploaded
            writefunc ([Dict[str, Any], filelike]): function used to write data to the file opened
        Returns:
            List[str]: List of GCS URLs corresponding to the uploaded frame batches
        """

        T = TypeVar("T")

        def batches(lst: List[T], size: int) -> Iterator[List[T]]:
            for i in range(0, len(lst), size):
                yield lst[i : i + size]

        def save_batch(assets: List[Dict[str, Any]]) -> Optional[str]:
            current_time = datetime.datetime.now()
            # Write frames to local temp file
            temp_frame_prefix = "al_{}_{}s_{}_".format(
                current_time.strftime("%Y%m%d_%H%M%S_%f"), asset_type, frame_batch_uuid
            )
            temp_frame_path = self._save_content_to_temp(
                temp_frame_prefix,
                lambda x: writefunc([asset for asset in assets], x),
            )
            return temp_frame_path

        # Get upload / download URLs
        download_urls = []
        get_upload_path = (
            f"{self.api_endpoint}/collection_campaigns/{campaign_id}/get_upload_url"
        )

        batch_filepaths = [
            save_batch(batched_assets)
            for batched_assets in batches(assets, MAX_FRAMES_PER_BATCH)
        ]

        # TODO: Review, since this is a real code change. Do we expect these to ever
        # be falsey?
        successful_filepaths = [x for x in batch_filepaths if x]

        file_ext = ".jsonl" if asset_type == "sample_frame" else ".arrow"

        download_urls = _upload_local_files(
            successful_filepaths,
            get_upload_path,
            self._get_creds_headers(),
            f"{asset_type}s_{frame_batch_uuid}",
            file_ext,
            delete_after_upload=True,
        )

        return download_urls

    # TODO: Is it intentional to return something here?
    def _post_collection_frames(self, campaign: Dict[str, Any]) -> Any:
        """takes a payload and posts it to the API

        Args:
            payload (Dict[str, Any]): Dict structure containing the payload for a campaign to post
        """
        print(f"Saving collection frames for {self._campaign_str(campaign)}...")

        payload = {
            "collection_campaign_id": campaign["id"],
            "issue_uuid": campaign["issue_uuid"],
            "dataframe_urls": self.camp_id_to_gcs_frame_urls[campaign["id"]],
            "num_processed": self.num_frames_processed,
            "embedding_urls": self.camp_id_to_gcs_embedding_urls[campaign["id"]],
        }

        r = requests_retry.post(
            self.api_endpoint + "/projects/blah/collection_frames",
            headers=self._get_creds_headers(),
            json=payload,
        )

        raise_resp_exception_error(r)
        return r.json()

    def _is_postprocessing_complete(self, campaign_summary: Dict[str, Any]) -> bool:
        return bool(
            campaign_summary.get("pca_signed_url")
            and campaign_summary.get("microcluster_info_signed_url")
        )

    def sync_state(
        self, target_project_names: List[str] = [], target_issue_uuids: List[str] = []
    ) -> None:
        """Downloads all collection campaigns and preps sampler with sample frames

        Args:
            target_project_names (Optional[List[str]], optional): List of project names whose collection campaigns should be sampled for
            target_issue_uuids (Optional[List[str]], optional): List of issue uuids whose collection campaigns should be sampled for. This takes precedence over 'target_project_names' and is mutually exclusive.
        """
        print("Starting Sync")
        all_coll_camps = self._get_all_campaigns(
            target_project_names=target_project_names,
            target_issue_uuids=target_issue_uuids,
        )

        # Skip over collection campaigns that still need to be post-processed
        processing_coll_camps_issue_uuids = [
            c.get("issue_uuid")
            for c in all_coll_camps
            if c["active"] and not self._is_postprocessing_complete(c)
        ]
        if len(processing_coll_camps_issue_uuids) > 0:
            print(
                f"{len(processing_coll_camps_issue_uuids)} collection campaigns still awaiting postprocessing."
            )
            for issue_uuid in processing_coll_camps_issue_uuids:
                print(f" - issue uuid: {issue_uuid}")

        self.active_coll_camp_summaries = {
            c["id"]: c
            for c in all_coll_camps
            if c["active"] and self._is_postprocessing_complete(c)
        }
        print(
            f"Found {len(self.active_coll_camp_summaries)} Active Collection Campaigns"
        )

        # Initialize a sampling agent per active, postprocessed campaign
        self.camp_id_to_sample_agent = {
            c["id"]: self.sampling_agent()
            for c in self.active_coll_camp_summaries.values()
        }

        print(f"Downloading assets for each Collection Campaign")
        for campaign in self.active_coll_camp_summaries.values():
            # download each of the preprocessed files for the example dataset locally
            signed_urls = {
                "pca_signed_url": campaign.get("pca_signed_url"),
                "microcluster_info_signed_url": campaign.get(
                    "microcluster_info_signed_url"
                ),
            }

            url_key_to_downloaded_file_path: Dict[str, Optional[str]] = {}
            for url_key, signed_url in signed_urls.items():
                if signed_url is None:
                    url_key_to_downloaded_file_path[url_key] = None
                    continue
                current_time = datetime.datetime.now()
                random_uuid = str(uuid4())
                temp_file_prefix = "al_{}_{}_{}_{}".format(
                    current_time.strftime("%Y%m%d_%H%M%S_%f"),
                    str(campaign.get("id")),
                    url_key,
                    random_uuid,
                )
                file_path = self._save_content_to_temp(
                    temp_file_prefix,
                    lambda x: self.download_to_file(signed_url, x),
                    mode="wb",
                )
                url_key_to_downloaded_file_path[url_key] = file_path

            path_key_to_downloaded_file_path = {
                k[:-10] + "path": v for k, v in url_key_to_downloaded_file_path.items()
            }

            # Load data into agent for each campaign
            agent = self.camp_id_to_sample_agent[campaign.get("id")]
            agent.load_sampling_dataset(
                element_type=campaign.get("element_type"),
                preprocessed_info=path_key_to_downloaded_file_path,
                primary_task=campaign.get("primary_task"),
            )

    def sample_probabilities(self, frames: List[LabeledFrame]) -> Any:
        """Takes a list of Labeled Frames and stores scores for each based on each synced collection campaigns

        Args:
            frames (List[LabeledFrame]): a List of Labeled frames to score based on synced Collection Campaigns
        """
        print("Sampling frames...")

        batch_uuid = str(uuid4())
        self.frame_batch_uuid_to_camp_id_to_probability_score_info[batch_uuid] = {}

        for campaign in self.active_coll_camp_summaries.values():
            print(f"Scoring frames for {self._campaign_str(campaign)}...")
            self.frame_batch_uuid_to_camp_id_to_probability_score_info[batch_uuid][
                campaign["id"]
            ] = [
                # A dict with fields (similarity_score, similarity_score_version, sampled_element_id, closest_microcluster_idx)
                self.camp_id_to_sample_agent[campaign["id"]].score_frame(frame)
                for frame in tqdm(frames, desc="Num frames scored")
            ]

        current_time = datetime.datetime.now()
        temp_frame_prefix = "al_{}_collection_campaign_candidate_frames_{}_".format(
            current_time.strftime("%Y%m%d_%H%M%S_%f"), batch_uuid
        )

        frame_dicts = []
        for frame in frames:
            frame_dict = frame.to_dict()
            frame_dict["label_data"] = frame._labels.to_dict()["label_data"]
            frame_dicts.append(frame_dict)

        frame_path = self._save_content_to_temp(
            temp_frame_prefix,
            lambda x: self._write_frames_to_file(frame_dicts, x),
        )
        self.frame_batch_uuid_to_temp_file_path[batch_uuid] = frame_path

        print("Sampling complete!")

        self.num_frames_processed += len(frames)

        return self.frame_batch_uuid_to_camp_id_to_probability_score_info[batch_uuid]

    def _upload_frames_and_embeddings_for_campaign(
        self,
        frame_batch_uuid: str,
        campaign: Dict[str, Any],
        frames: List[Dict[str, Any]],
        embeddings: List[Dict[str, Any]],
    ) -> None:
        if len(frames) == 0:
            return

        print(f"Uploading Frames for {self._campaign_str(campaign)}")
        dataframe_urls = self._upload_collection_assets_to_gcs(
            frame_batch_uuid,
            campaign["id"],
            frames,
            self._write_frames_to_file,
            "sample_frame",
        )

        print(f"Uploading Embeddings for {self._campaign_str(campaign)}")
        embedding_urls = self._upload_collection_assets_to_gcs(
            frame_batch_uuid,
            campaign["id"],
            embeddings,
            self._write_embeddings_to_file,
            "embedding",
        )

        if campaign["id"] not in self.camp_id_to_gcs_frame_urls:
            self.camp_id_to_gcs_frame_urls[campaign["id"]] = dataframe_urls
            self.camp_id_to_gcs_embedding_urls[campaign["id"]] = embedding_urls
        else:
            self.camp_id_to_gcs_frame_urls[campaign["id"]].extend(dataframe_urls)
            self.camp_id_to_gcs_embedding_urls[campaign["id"]].extend(embedding_urls)

    # Based on Welford's algorithm for updating mean + variance
    def _update_frame_statistics(self, similarity_score: float) -> None:
        self.sampled_frame_statistics["n"] += 1

        delta = similarity_score - self.sampled_frame_statistics["mean"]
        self.sampled_frame_statistics["mean"] += (
            delta / self.sampled_frame_statistics["n"]
        )
        delta2 = similarity_score - self.sampled_frame_statistics["mean"]
        self.sampled_frame_statistics["m2"] += delta * delta2
        self.sampled_frame_statistics["variance"] = (
            self.sampled_frame_statistics["m2"] / self.sampled_frame_statistics["n"]
        )

        if similarity_score < self.sampled_frame_statistics["min"]:
            self.sampled_frame_statistics["min"] = similarity_score
        if similarity_score > self.sampled_frame_statistics["max"]:
            self.sampled_frame_statistics["max"] = similarity_score

    def _save_for_collection_with_threshold(
        self, override_sampling_threshold: float = None, dry_run: bool = False
    ) -> None:
        num_frame_batches = len(self.frame_batch_uuid_to_temp_file_path)

        for idx, frame_batch_uuid in enumerate(
            self.frame_batch_uuid_to_temp_file_path.keys()
        ):
            print(f"Processing frame batch {idx + 1} of {num_frame_batches}...")

            frames = self._read_rows_from_disk(
                self.frame_batch_uuid_to_temp_file_path[frame_batch_uuid]
            )
            camp_id_to_probability_score_info = (
                self.frame_batch_uuid_to_camp_id_to_probability_score_info[
                    frame_batch_uuid
                ]
            )
            for campaign in self.active_coll_camp_summaries.values():
                scores = camp_id_to_probability_score_info[campaign["id"]]
                sampling_threshold = (
                    DEFAULT_SAMPLING_THRESHOLD
                    if campaign.get("sampling_threshold") is None
                    else campaign.get("sampling_threshold")
                )
                sampling_threshold = (
                    sampling_threshold
                    if override_sampling_threshold is None
                    else override_sampling_threshold
                )
                filtered_frame_indexes_and_scores = filter(
                    lambda score: score[1].get("similarity_score")
                    >= sampling_threshold,
                    enumerate(scores),
                )
                filtered_frame_indexes = map(
                    lambda score: score[0], filtered_frame_indexes_and_scores
                )

                filtered_frames_dict = []
                sample_embeddings: List[Dict[str, Any]] = []
                for idx in filtered_frame_indexes:
                    frame_dict = frames[idx]
                    # Add info from scoring (e.g. similarity score)
                    frame_dict["similarity_score"] = scores[idx].get("similarity_score")
                    frame_dict["sampled_element_id"] = scores[idx].get(
                        "sampled_element_id"
                    )
                    frame_dict["similarity_score_version"] = scores[idx].get(
                        "similarity_score_version"
                    )
                    frame_dict["closest_microcluster_idx"] = scores[idx].get(
                        "closest_microcluster_idx"
                    )
                    frame_dict["model_id"] = scores[idx].get("model_id")

                    filtered_frames_dict.append(frame_dict)
                    sample_embeddings.append(
                        {
                            "task_id": frame_dict.get("task_id"),
                            "embedding": scores[idx].get("elt_embedding_vec"),
                        }
                    )

                    # Update aggregate statistics
                    if (
                        frame_dict["similarity_score"]
                        > self.sampled_frame_statistics["max"]
                    ):
                        self.most_similar_frame = frame_dict
                        self.most_similar_frame_campaign_id = campaign["id"]
                    self._update_frame_statistics(frame_dict["similarity_score"])

                if not dry_run:
                    self._upload_frames_and_embeddings_for_campaign(
                        frame_batch_uuid,
                        campaign,
                        filtered_frames_dict,
                        sample_embeddings,
                    )

    def _save_for_collection_with_count(
        self, target_sample_count: int, dry_run: bool = False
    ) -> None:
        print(f"Targeting sample frame count of {target_sample_count}...")

        # If we're targeting a specific frame count, keep a global queue across batches
        # The extra slot allows us to determine what to keep when the queue reaches target_sample_count
        frame_queue: "PriorityQueue[Tuple[float, str, str, Dict[str, Any], List[float]]]" = PriorityQueue(
            target_sample_count + 1
        )

        with tqdm(
            total=self.num_frames_processed,
            file=sys.stdout,
            unit_scale=True,
            desc="Frame Processing Progress",
        ) as pbar:  # type: tqdm[Any]
            for idx, frame_batch_uuid in enumerate(
                self.frame_batch_uuid_to_temp_file_path.keys()
            ):
                frames = self._read_rows_from_disk(
                    self.frame_batch_uuid_to_temp_file_path[frame_batch_uuid]
                )
                camp_id_to_probability_score_info = (
                    self.frame_batch_uuid_to_camp_id_to_probability_score_info[
                        frame_batch_uuid
                    ]
                )

                for frame_idx, frame in enumerate(frames):
                    max_score_embedding: List[float] = []
                    max_score_campaign_id: str = ""
                    max_similarity_score: float = -1

                    # Determine which campaign this frame is the best match for
                    for campaign in self.active_coll_camp_summaries.values():
                        score_info = camp_id_to_probability_score_info[campaign["id"]][
                            frame_idx
                        ]
                        if score_info.get("similarity_score") > max_similarity_score:
                            max_similarity_score = score_info.get("similarity_score")
                            max_score_campaign_id = campaign["id"]
                            # Extract the matching embedding info
                            max_score_embedding = score_info.get("elt_embedding_vec")
                            score_info = {
                                k: v
                                for k, v in score_info.items()
                                if k != "elt_embedding_vec"
                            }
                            # Then merge score_info into the frame_dict
                            frame.update(score_info)

                    # No matching embedding was found for this frame
                    # (e.g. frame itself might not have embeddings)
                    if frame.get("sampled_element_id") is None:
                        pbar.update(1)
                        continue

                    frame_task_id: str = frame["task_id"]
                    frame_queue.put(
                        (
                            max_similarity_score,
                            frame_task_id,  # To help with uniqueness/sorting
                            max_score_campaign_id,
                            frame,
                            max_score_embedding,
                        )
                    )

                    # If frame queue is full, we need to remove the lowest to keep room open for sorting the next entry
                    if frame_queue.full():
                        frame_queue.get()

                    pbar.update(1)

        campaign_id_to_frames: Dict[str, List[Dict[str, Any]]] = {
            campaign["id"]: [] for campaign in self.active_coll_camp_summaries.values()
        }
        campaign_id_to_embeddings: Dict[str, List[Dict[str, Any]]] = {
            campaign["id"]: [] for campaign in self.active_coll_camp_summaries.values()
        }

        while not frame_queue.empty():
            score, task_id, campaign_id, frame_dict, embedding = frame_queue.get()

            # Update aggregate statistics
            if frame_dict["similarity_score"] > self.sampled_frame_statistics["max"]:
                self.most_similar_frame = frame_dict
                self.most_similar_frame_campaign_id = campaign_id

            self._update_frame_statistics(frame_dict["similarity_score"])

            campaign_id_to_frames[campaign_id].append(frame_dict)
            campaign_id_to_embeddings[campaign_id].append(
                {
                    "task_id": task_id,
                    "embedding": embedding,
                }
            )

        if not dry_run:
            # We need a filler uuid (since we're not uploading by frame batch)
            filler_uuid = str(uuid4())
            for campaign in self.active_coll_camp_summaries.values():
                self._upload_frames_and_embeddings_for_campaign(
                    filler_uuid,
                    campaign,
                    campaign_id_to_frames[campaign["id"]],
                    campaign_id_to_embeddings[campaign["id"]],
                )

    def _report_sampled_frame_statistics(self) -> None:
        print("===================")
        print("Stats Summary for Sampled Frames")
        print("===================")
        print(
            f"- Num collected samples / total num frames seen: {self.sampled_frame_statistics['n']} / {self.num_frames_processed}"
        )
        print(f"- Min similarity score: {self.sampled_frame_statistics['min']:.3f}")
        print(f"- Max similarity score: {self.sampled_frame_statistics['max']:.3f}")
        print(f"- Mean similarity score: {self.sampled_frame_statistics['mean']:.3f}")
        print(
            f"- Standard deviation similarity score: {math.sqrt(self.sampled_frame_statistics['variance']):.3f}\n"
        )

    def save_for_collection(
        self,
        override_sampling_threshold: float = None,
        target_sample_count: int = None,
        dry_run: bool = False,
    ) -> None:
        """Based on the sampling threshold, take all sampled frames and upload those that score above
        the sampling threshold for each Collection Campaign.

        Args:
            override_sampling_threshold (float, optional): Override sampling threshold for all campaigns to save to server. If this value is set it will not use the sampling threshold set in UI for each Collection Campaign.
            target_sample_count(int, optional): Instead of collecting all samples above a specified similarity threshold, you can specify this option to save the N most similar samples.
        """

        if dry_run:
            print("===================")
            print("DRY RUN: No collection frames will be uploaded.")
            print("===================")

        # First determine which frames/embeddings to save and upload to GCS
        if target_sample_count is not None:
            self._save_for_collection_with_count(target_sample_count, dry_run=dry_run)
        else:
            self._save_for_collection_with_threshold(
                override_sampling_threshold, dry_run=dry_run
            )

        self._report_sampled_frame_statistics()

        if dry_run:
            if (
                self.most_similar_frame_campaign_id is not None
                and self.most_similar_frame is not None
            ):
                print("===================")
                print("Preview Frame")
                print("===================")
                print("Generating preview frame for highest scoring sample...")
                print(
                    f"Similarity score: {self.most_similar_frame['similarity_score']:.3f}\n"
                )
                # Preview first frame
                preview_frame_campaign: Dict[
                    str, Any
                ] = self.active_coll_camp_summaries[self.most_similar_frame_campaign_id]

                self._preview_frame_dict(
                    project_id=preview_frame_campaign["project_name"],
                    both_frames_dict={
                        "labeled_frame": self.most_similar_frame,
                        "inference_frame": None,
                    },
                    selected_label=self.most_similar_frame["sampled_element_id"]
                    if preview_frame_campaign["element_type"] == "crop"
                    else None,
                )
        else:
            # Now create the collection frames in the DB
            for campaign in self.active_coll_camp_summaries.values():
                if campaign["id"] in self.camp_id_to_gcs_frame_urls:
                    resp = self._post_collection_frames(campaign)
                    print(
                        "Successfully saved {} new frames for {} ({} frames already referenced by existing samples)".format(
                            resp.get("num_new_frames"),
                            self._campaign_str(campaign),
                            resp.get("num_duplicate_frames"),
                        )
                    )
                else:
                    print(f"No samples matched for {self._campaign_str(campaign)}")

            mark_temp_directory_complete(self.temp_file_path)
            self._clear_cached_sample_state()



"""frames.py
============
The base frame and update frame classes.
"""

import datetime
import json
import numbers
from io import IOBase
from tempfile import NamedTemporaryFile
import pyarrow as pa
import numpy as np
import pandas as pd
import re
from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, cast
from typing_extensions import Protocol, TypedDict

from .util import (
    ALL_CAMERA_MODELS,
    ALL_ORIENTATION_KEYS,
    ALL_POSITION_KEYS,
    CAMERA_MODELS,
    DEFAULT_SENSOR_ID,
    ResizeMode,
    _is_one_gb_available,
    assert_and_get_valid_iso_date,
    assert_valid_name,
    add_object_user_attrs,
    create_temp_directory,
    is_valid_float,
    is_valid_number,
    mark_temp_directory_complete,
    extract_illegal_characters,
    TYPE_PRIMITIVE_TO_STRING_MAP,
    USER_METADATA_SEQUENCE,
    USER_METADATA_PRIMITIVE_TYPES,
    USER_METADATA_MODE_TYPES,
    SUPPORTED_USER_METADATA_TYPES,
    POLYGON_VERTICES_KEYS,
    POSITION_KEYS,
    ORIENTATION_KEYS,
    KEYPOINT_KEYS,
    MAX_FRAMES_PER_BATCH,
    GtLabelEntryDict,
    FrameEmbeddingDict,
    CropEmbeddingDict,
    LabelType,
    LabelFrameSummary,
    InferenceFrameSummary,
    UpdateType,
)

from .labels import GTLabelSet, InferenceLabelSet

# TODO: optional types aren't supported in python, so we should avoid it
# TODO: Give these explicit, concrete typing
CoordinateFrameDict = Dict[str, Any]
SensorDataDict = Dict[str, Any]
GeoDataDict = Dict[str, Any]
UserMetadataEntry = Tuple[
    str,
    SUPPORTED_USER_METADATA_TYPES,
    USER_METADATA_PRIMITIVE_TYPES,
    USER_METADATA_MODE_TYPES,
]


# TODO: handle diffs?
class BaseFrame:
    """A labeled frame for a dataset.

    Args:
        frame_id (str): A unique id for this frame.
        date_captured (str, optional): ISO formatted datetime string. Defaults to None.
        device_id (str, optional): The device that generated this frame. Defaults to None.
    """

    device_id: str
    date_captured: str
    frame_id: str
    coordinate_frames: List[CoordinateFrameDict]
    sensor_data: List[SensorDataDict]
    geo_data: GeoDataDict
    user_metadata: List[UserMetadataEntry]
    _coord_frame_ids_set: Set[str]
    embedding: Optional[FrameEmbeddingDict]
    _label_ids_set: Set[str]
    reuse_latest_embedding: Optional[bool]
    frame_embedding_dim: int
    update_type: UpdateType

    def __init__(
        self,
        *,
        frame_id: str,
        date_captured: Optional[str] = None,
        device_id: Optional[str] = None,
        reuse_latest_embedding: Optional[bool] = False,
        update_type: UpdateType = "ADD",
    ) -> None:
        if not isinstance(frame_id, str):
            raise Exception("frame ids must be strings")

        illegal_frame_id_characters = extract_illegal_characters(frame_id)
        if illegal_frame_id_characters:
            raise Exception(
                f"frame ids cannot contain the following character(s): {illegal_frame_id_characters}"
            )

        self.frame_id = frame_id

        if date_captured is not None:
            self.date_captured = assert_and_get_valid_iso_date(date_captured)
        else:
            self.date_captured = str(datetime.datetime.now())

        if device_id is not None:
            self.device_id = device_id
        else:
            self.device_id = "default_device"

        self.coordinate_frames = []
        self.sensor_data = []
        self.geo_data = {}
        self.user_metadata = []
        self.embedding = None
        self.reuse_latest_embedding = reuse_latest_embedding
        self.frame_embedding_dim = -1

        self._coord_frame_ids_set = set()
        self._label_ids_set = set()
        self.update_type = update_type

    def _add_coordinate_frame(self, coord_frame_obj: Dict[str, str]) -> None:
        """Add coordinate frame to the dataset frame

        Args:
            coord_frame_obj (Dict[str, str]): takes in 'coordinate_frame_id', 'coordinate_frame_type' and optional 'coordinate_frame_metadata'(json dict)
        """
        self.coordinate_frames.append(coord_frame_obj)
        self._coord_frame_ids_set.add(coord_frame_obj["coordinate_frame_id"])

    def _coord_frame_exists(self, coord_frame_id: str) -> bool:
        """Check to see if the coord frame id is already part of the frame

        Args:
            coord_frame_id (str): The coord frame id to check for inclusion

        Returns:
            bool: whether or not the coord frame id is in the frame set
        """
        return coord_frame_id in self._coord_frame_ids_set

    def add_frame_embedding(
        self, *, embedding: List[float], model_id: str = ""
    ) -> None:
        """Add an embedding to this frame

        Args:
            embedding (List[float]): A vector of floats between length 0 and 12,000.
            model_id (str, optional): The model id used to generate these embeddings. Defaults to "".
        """
        if not embedding or len(embedding) > 12000:
            raise Exception("Length of embeddings should be between 0 and 12,000")

        for embedding_el in embedding:
            if not isinstance(embedding_el, numbers.Number):
                raise Exception(
                    f"Unexpectedly encountered a {type(embedding[0])} element. Only flat arrays of numbers are supported for embeddings."
                )

        if not self.embedding:
            self.frame_embedding_dim = len(embedding)
            self.embedding = {
                "task_id": self.frame_id,
                "model_id": model_id,
                "date_generated": str(datetime.datetime.now()),
                "embedding": embedding,
            }
            self.reuse_latest_embedding = False

    def add_user_metadata(
        self,
        key: str,
        val: Union[str, int, float, bool],
        val_type: Optional[USER_METADATA_PRIMITIVE_TYPES] = None,
    ) -> None:
        """Add a user provided metadata field.

        The types of these metadata fields will be infered and they'll be made
        available in the app for querying and metrics.

        Args:
            key (str): The key for your metadata field
            val (Union[str, int, float, bool]): value
            val_type (Literal["str", "int", "float", "bool"], optional): type of val as string. Defaults to None.
        """
        assert_valid_name(key)
        # Validates that neither val or type is None
        if val is None and val_type is None:
            raise Exception(
                f"For frame_id {self.frame_id}: User Metadata key {key} must provide "
                f"scalar value or expected type of scalar value if None"
            )
        # Validates that val has an accepted type
        if val is not None and type(val) not in TYPE_PRIMITIVE_TO_STRING_MAP:
            raise Exception(
                f"For frame_id {self.frame_id}: User Metadata Value {val} "
                f"not in accepted scalar value types {TYPE_PRIMITIVE_TO_STRING_MAP.values()}"
            )
        # Validates that val_type has an accepted type
        if val_type and val_type not in TYPE_PRIMITIVE_TO_STRING_MAP.values():
            raise Exception(
                f"For frame_id {self.frame_id}: User Metadata Value Type {val_type} "
                f"not in accepted scalar value types {TYPE_PRIMITIVE_TO_STRING_MAP.values()}"
            )

        # Sets val_type if it is not set
        if val_type is None:
            val_type = TYPE_PRIMITIVE_TO_STRING_MAP[type(val)]

        # If type is float, ensure that it is a valid float
        if val is not None and val_type == "float" and not is_valid_float(val):
            raise Exception(
                f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
                f"type is inferred as float but is not a valid float"
            )

        # Checks that inferred type matches what the user put in val_type
        if val is not None:
            for (
                primitive,
                type_string,
            ) in TYPE_PRIMITIVE_TO_STRING_MAP.items():
                if type(val) is primitive and val_type != type_string:
                    raise Exception(
                        f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
                        f"type is inferred as {type_string} but provided type was {val_type}"
                    )

        self.user_metadata.append((key, val, val_type, "scalar"))

    def add_user_metadata_list(
        self,
        key: str,
        val: USER_METADATA_SEQUENCE,
        list_elt_type: Optional[USER_METADATA_PRIMITIVE_TYPES] = None,
    ) -> None:
        """Add a user provided metadata list field.

        The types of these metadata fields will be infered and they'll be made
        available in the app for querying and metrics.

        Args:
            key (str): The key for your metadata field
            val (Union[List[int], List[str], List[bool], List[float], Tuple[int], Tuple[str], Tuple[bool], Tuple[float]]): value
            list_elt_type (Literal["str", "int", "float", "bool"], optional): type of list elements as string. Applies to all members of val. Defaults to None.
        """
        assert_valid_name(key)
        # Validates that neither val or type or mode is None
        if val is None and list_elt_type is None:
            raise Exception(
                f"For frame_id {self.frame_id}: User Metadata list key {key} must provide "
                f"list or expected type of list elements if None"
            )

        # Validates that val has an accepted type
        if val is not None:
            if type(val) not in (list, tuple):
                raise Exception(
                    f"For frame_id {self.frame_id}: User Metadata list value {val} "
                    f"is not in accepted sequence types list, tuple."
                )
            if len(val) == 0 and not list_elt_type:
                raise Exception(
                    f"For frame_id {self.frame_id}: User Metadata list value {val} "
                    f"is an empty {type(val).__name__}. "
                    "Please specify the expected scalar value type for its elements by passing a list_elt_type"
                )

            # validate that all elements in the list are the same primitive type, based on the first element
            if len(val) > 0:
                found_val_types = {type(el) for el in val}

                if len(found_val_types) > 1:
                    raise Exception(
                        f"For frame_id {self.frame_id}: User Metadata list value {val} "
                        f"has elements of invalid type. Expected all elements to be of the same type, found types of {found_val_types}"
                    )

                inferred_val_type = found_val_types.pop()
                if inferred_val_type not in TYPE_PRIMITIVE_TO_STRING_MAP:
                    raise Exception(
                        f"For frame_id {self.frame_id}: User Metadata list value {val} contains elements "
                        f"not in accepted scalar value types {TYPE_PRIMITIVE_TO_STRING_MAP.values()}"
                    )

        # Validates that list_elt_type has an accepted type
        if list_elt_type and list_elt_type not in TYPE_PRIMITIVE_TO_STRING_MAP.values():
            raise Exception(
                f"For frame_id {self.frame_id}: User Metadata list element type {list_elt_type} "
                f"not in accepted scalar value types {TYPE_PRIMITIVE_TO_STRING_MAP.values()}"
            )

        # Checks that inferred type matches what the user put in list_elt_type
        if list_elt_type is not None and val is not None and len(val) > 0:
            inferred_val_type = type(val[0])
            for (
                primitive,
                type_string,
            ) in TYPE_PRIMITIVE_TO_STRING_MAP.items():
                if inferred_val_type is primitive and list_elt_type != type_string:
                    raise Exception(
                        f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
                        f"element type is inferred as {type_string} but provided type was {list_elt_type}"
                    )

        # Sets list_elt_type if it is not set
        if list_elt_type is None:
            inferred_val_type = type(val[0])
            list_elt_type = TYPE_PRIMITIVE_TO_STRING_MAP[inferred_val_type]

        # If element type is float, ensure that all elements are valid floats
        if val is not None and list_elt_type == "float":
            for val_elt in val:
                if not is_valid_float(val_elt):
                    raise Exception(
                        f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
                        f"element type is inferred as float but is not a valid float"
                    )

        self.user_metadata.append((key, val, list_elt_type, "list"))

    def add_geo_latlong_data(self, lat: float, lon: float) -> None:
        """Add a user provided EPSG:4326 WGS84 lat long pair to each frame

        We expect these values to be floats

        Args:
            lat (float): lattitude of Geo Location
            lon (float): longitude of Geo Location
        """
        if not (isinstance(lat, float) and isinstance(lon, float)):
            raise Exception(
                f"Lattitude ({lat}) and Longitude ({lon}) must both be floats."
            )

        self.geo_data["geo_EPSG4326_lat"] = lat
        self.geo_data["geo_EPSG4326_lon"] = lon

    def add_point_cloud_pcd(
        self,
        *,
        sensor_id: str,
        pcd_url: str,
        coord_frame_id: Optional[str] = None,
        date_captured: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add a point cloud sensor data point to this frame,
        contained in PCD format. ascii, binary, and binary_compressed formats are supported.
        Numeric values for the following column names are expected:
        x, y, z, intensity (optional), range (optional)

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            pcd_url (str): URL to PCD formated data
            coord_frame_id (Optional[str], optional): The coordinate frame id. Defaults to None.
            date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed url assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        if coord_frame_id is None:
            coord_frame_id = "world"

        data_urls = {
            "pcd_url": pcd_url,
        }

        if not self._coord_frame_exists(coord_frame_id):
            if coord_frame_id == "world":
                self._add_coordinate_frame(
                    {
                        "coordinate_frame_id": coord_frame_id,
                        "coordinate_frame_type": "WORLD",
                    }
                )
            else:
                raise Exception(
                    "Coordinate frame {} does not exist.".format(coord_frame_id)
                )

        self.sensor_data.append(
            {
                "coordinate_frame": coord_frame_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": {},
                "sensor_type": "POINTCLOUD_PCD_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_point_cloud_bins(
        self,
        *,
        sensor_id: str,
        pointcloud_url: str = None,
        kitti_format_url: str = None,
        intensity_url: str = None,
        range_url: str = None,
        coord_frame_id: Optional[str] = None,
        date_captured: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add a point cloud sensor data point to this frame, contained in dense binary files of
        little-endian values, similar to the raw format of KITTI lidar data. You can provide
        a combination of the following values, as long as at least either kitti_format_url or
        pointcloud_url are provided.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            pointcloud_url (Optional[str]): URL for the pointcloud: float32 [x1, y1, z1, x2, y2, z2, ...]
            kitti_format_url (Optional[str]): URL for the pointcloud + intensity: float32 [x1, y1, z1, i1, x2, y2, z2, i2, ...]
            intensity_url (Optional[str]): URL for the Intensity Pointcloud: unsigned int32 [i1, i2, ...]
            range_url (Optional[str]): URL for the Range Pointcloud: float32 [r1, r2, ...]
            coord_frame_id (Optional[str], optional): Id for the Coordinate Frame. Defaults to None.
            date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        if pointcloud_url is None and kitti_format_url is None:
            raise Exception(
                "Either pointcloud_url or kitti_format_url must be provided."
            )

        if pointcloud_url is not None and kitti_format_url is not None:
            raise Exception(
                "Only one of pointcloud_url or kitti_format_url must be provided."
            )

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        if coord_frame_id is None:
            coord_frame_id = "world"

        data_urls = {}
        if pointcloud_url is not None:
            data_urls["pointcloud_url"] = pointcloud_url
        if kitti_format_url is not None:
            data_urls["kitti_format_url"] = kitti_format_url
        if range_url is not None:
            data_urls["range_url"] = range_url
        if intensity_url is not None:
            data_urls["intensity_url"] = intensity_url

        if not self._coord_frame_exists(coord_frame_id):
            if coord_frame_id == "world":
                self._add_coordinate_frame(
                    {
                        "coordinate_frame_id": coord_frame_id,
                        "coordinate_frame_type": "WORLD",
                    }
                )
            else:
                raise Exception(
                    "Coordinate frame {} does not exist.".format(coord_frame_id)
                )

        self.sensor_data.append(
            {
                "coordinate_frame": coord_frame_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": {},
                "sensor_type": "POINTCLOUD_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_obj(
        self,
        *,
        sensor_id: str,
        obj_url: str,
        coord_frame_id: Optional[str] = None,
        date_captured: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add a .obj file to the frame for text based geometry

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            obj_url (str): URL to where the object is located
            coord_frame_id (Optional[str], optional): ID for the coordinate frame. Defaults to None.
            date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        if coord_frame_id is None:
            coord_frame_id = "world"

        data_urls = {
            "obj_url": obj_url,
        }

        if not self._coord_frame_exists(coord_frame_id):
            self._add_coordinate_frame(
                {
                    "coordinate_frame_id": coord_frame_id,
                    "coordinate_frame_type": "WORLD",
                }
            )

        self.sensor_data.append(
            {
                "coordinate_frame": coord_frame_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": {},
                "sensor_type": "OBJ_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_image(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        image_url: str,
        preview_url: Optional[str] = None,
        date_captured: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        coord_frame_id: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add an image "sensor" data to this frame.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            image_url (str): The URL to load this image data.
            preview_url (Optional[str], optional): A URL to a compressed version of the image. It must be the same pixel dimensions as the original image. Defaults to None.
            date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
            width (Optional[int], optional): The width of the image in pixels. Defaults to None.
            height (Optional[int], optional): The height of the image in pixels. Defaults to None.
            coord_frame_id (Optional[str], optional): Id for the Coordinate Frame (only accepts coordinate frames of type "IMAGE" or "IMAGE_PROJECTION"). Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        sensor_metadata = {}
        if width is not None:
            if not isinstance(width, int):
                raise Exception("width must be an int")
            sensor_metadata["width"] = width

        if height is not None:
            if not isinstance(height, int):
                raise Exception("height must be an int")
            sensor_metadata["height"] = height

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        data_urls = {"image_url": image_url}
        if preview_url is not None:
            data_urls["preview_url"] = preview_url

        if coord_frame_id is None:
            coord_frame_id = sensor_id
            self._add_coordinate_frame(
                {
                    "coordinate_frame_id": coord_frame_id,
                    "coordinate_frame_type": "IMAGE",
                }
            )
        elif not self._coord_frame_exists(coord_frame_id):
            raise Exception(
                "Coordinate frame {} does not exist.".format(coord_frame_id)
            )

        for coordinate_frame in self.coordinate_frames:
            coord_type = coordinate_frame["coordinate_frame_type"]
            if (
                coordinate_frame["coordinate_frame_id"] == coord_frame_id
                and coord_type != "IMAGE"
                and coord_type != "IMAGE_PROJECTION"
            ):
                raise Exception(
                    'Coordinate frame {} is of type {} and must be either "IMAGE" or "IMAGE_PROJECTION".'.format(
                        coord_frame_id, coord_type
                    )
                )

        self.sensor_data.append(
            {
                "coordinate_frame": coord_frame_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": sensor_metadata,
                "sensor_type": "IMAGE_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_image_overlay(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        overlay_id: str,
        image_url: str,
        date_captured: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        mirror_asset: Optional[bool] = False,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """Add an image overlay for the given "sensor" to this frame.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            overlay_id (str): The id of this overlay.
            image_url (str): The URL to load this image data.
            date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
            width (Optional[int], optional): The width of the image in pixels. Defaults to None.
            height (Optional[int], optional): The height of the image in pixels. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
            resize_mode (ResizeMode, optional): if the overlay is a different size from the base image, define how to display it
                "fill": stretch the overlay to fit the base image dimensions
                None: no change
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        if not isinstance(overlay_id, str):
            raise Exception("overlay ids must be strings")

        sensor_metadata: Dict[str, Union[str, int]] = {
            "overlay_id": overlay_id,
            "resize_mode": str(resize_mode),
        }
        if width is not None:
            if not isinstance(width, int):
                raise Exception("width must be an int")
            sensor_metadata["width"] = width

        if height is not None:
            if not isinstance(height, int):
                raise Exception("height must be an int")
            sensor_metadata["height"] = height

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        data_urls = {"image_url": image_url}

        self._add_coordinate_frame(
            {"coordinate_frame_id": sensor_id, "coordinate_frame_type": "IMAGE"}
        )
        self.sensor_data.append(
            {
                "coordinate_frame": sensor_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": sensor_metadata,
                "sensor_type": "IMAGE_OVERLAY_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_audio(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        audio_url: str,
        date_captured: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add an audio "sensor" data to this frame.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            audio_url (str): The URL to load this audio data (mp3, etc.).
            date_captured (str, optional): ISO formatted date. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        sensor_metadata: Dict[str, Any] = {}
        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        data_urls = {"audio_url": audio_url}

        self._add_coordinate_frame(
            {"coordinate_frame_id": sensor_id, "coordinate_frame_type": "AUDIO"}
        )
        self.sensor_data.append(
            {
                "coordinate_frame": sensor_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": sensor_metadata,
                "sensor_type": "AUDIO_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_video(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        video_url: str,
        date_captured: Optional[str] = None,
        mirror_asset: Optional[bool] = False,
    ) -> None:
        """Add a video "sensor" data to this frame.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            video_url (str): The URL to load this video data (mp4, webm, etc.).
            date_captured (str, optional): ISO formatted date. Defaults to None.
            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
        """
        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        sensor_metadata: Dict[str, Any] = {}
        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        date_captured = assert_and_get_valid_iso_date(date_captured)

        data_urls = {"video_url": video_url}

        self._add_coordinate_frame(
            {"coordinate_frame_id": sensor_id, "coordinate_frame_type": "VIDEO"}
        )
        self.sensor_data.append(
            {
                "coordinate_frame": sensor_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": sensor_metadata,
                "sensor_type": "VIDEO_V0",
                "mirror_asset": mirror_asset,
            }
        )

    def add_coordinate_frame_3d(
        self,
        *,
        coord_frame_id: str,
        position: Optional[Dict[POSITION_KEYS, Union[int, float]]] = None,
        orientation: Optional[Dict[ORIENTATION_KEYS, Union[int, float]]] = None,
        parent_frame_id: Optional[str] = None,
    ) -> None:
        """Add a 3D Coordinate Frame.

        Args:
            coord_frame_id (str): String identifier for this coordinate frame
            position (Optional[Dict[POSITION, Union[int, float]]], optional): Dict of the form {x, y, z}. Defaults to None.
            orientation (Optional[Dict[ORIENTATION, Union[int, float]]], optional): Quaternion rotation dict of the form {w, x, y, z}. Defaults to None.
            parent_frame_id (Optional[str], optional): String id of the parent coordinate frame. Defaults to None.
        """

        if not isinstance(coord_frame_id, str):
            raise Exception("coord_frame_id must be a string")

        if coord_frame_id == "world":
            raise Exception("coord_frame_id cannot be world")

        if self._coord_frame_exists(coord_frame_id):
            raise Exception("Coordinate frame already exists.")

        # If world doesn't exist, make the world coordinate frame
        if not self._coord_frame_exists("world"):
            self._add_coordinate_frame(
                {
                    "coordinate_frame_id": "world",
                    "coordinate_frame_type": "WORLD",
                }
            )

        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        if type(position) is not dict:
            raise Exception("position improperly formatted")
        for position_key in ALL_POSITION_KEYS:
            if not is_valid_number(position[position_key]):
                raise Exception("position coordinates must be valid numbers")

        if orientation is None:
            orientation = {"w": 1, "x": 0, "y": 0, "z": 0}
        if type(orientation) is not dict:
            raise Exception("orientation improperly formatted")
        for orientation_key in ALL_ORIENTATION_KEYS:
            if not is_valid_number(orientation[orientation_key]):
                raise Exception("orientation coordinates must be valid numbers")

        if parent_frame_id is None:
            parent_frame_id = "world"

        if not isinstance(parent_frame_id, str):
            raise Exception("parent_frame_id must be a string or None.")

        metadata = {
            "position": position,
            "orientation": orientation,
            "parent_frame_id": parent_frame_id,
        }

        self._add_coordinate_frame(
            {
                "coordinate_frame_id": coord_frame_id,
                "coordinate_frame_type": "WORLD",
                "coordinate_frame_metadata": json.dumps(metadata),
            }
        )

    def add_coordinate_frame_2d(
        self,
        *,
        coord_frame_id: str,
        fx: Union[int, float],
        fy: Union[int, float],
        camera_model: Optional[CAMERA_MODELS] = None,
        position: Optional[Dict[POSITION_KEYS, Union[int, float]]] = None,
        orientation: Optional[Dict[ORIENTATION_KEYS, Union[int, float]]] = None,
        camera_matrix: Optional[List[List[Union[int, float]]]] = None,
        k1: Optional[Union[int, float]] = None,
        k2: Optional[Union[int, float]] = None,
        k3: Optional[Union[int, float]] = None,
        k4: Optional[Union[int, float]] = None,
        k5: Optional[Union[int, float]] = None,
        k6: Optional[Union[int, float]] = None,
        p1: Optional[Union[int, float]] = None,
        p2: Optional[Union[int, float]] = None,
        cx: Optional[Union[int, float]] = None,
        cy: Optional[Union[int, float]] = None,
        s1: Optional[Union[int, float]] = None,
        s2: Optional[Union[int, float]] = None,
        s3: Optional[Union[int, float]] = None,
        s4: Optional[Union[int, float]] = None,
        skew: Optional[Union[int, float]] = None,
        parent_frame_id: Optional[str] = None,
    ) -> None:
        """Add a 2D Coordinate Frame.

        Args:
            coord_frame_id (str): String identifier for this coordinate frame.
            fx (Union[int, float]): focal length x in pixels.
            fy (Union[int, float]): focal length y in pixels.
            camera_model (Optional[CAMERA_MODELS], optional): Either "fisheye" for the fisheye model, or "brown_conrady" for the pinhole model with Brown-Conrady distortion. Defaults to "brown_conrady".
            position (Optional[Dict[POSITION, Union[int, float]]], optional): Dict of the form {x, y, z}. Defaults to {x: 0, y: 0, z: 0}.
            orientation (Optional[Dict[ORIENTATION, Union[int, float]]], optional): Quaternion rotation dict of the form {w, x, y, z}. Defaults to {x: 0, y: 0, z: 0, w: 1}.
            camera_matrix (Optional[List[List[Union[int, float]]]], optional): 4x4 row major order camera matrix mapping 3d world space to camera space (x right, y down, z forward). Keep in mind, if you pass in the camera matrix it will stack on top of the position/orientation you pass in as well. This is only needed if you cannot properly represent your camera using the position/orientation parameters. Defaults to None.
            cx (Optional[Union[int, float]], optional): optical center pixel x coordinate. Defaults to x center of image.
            cy (Optional[Union[int, float]], optional): optical center pixel y coordinate. Defaults to y center of image.
            k1 (Optional[Union[int, float]], optional): k1 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
            k2 (Optional[Union[int, float]], optional): k2 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
            k3 (Optional[Union[int, float]], optional): k3 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
            k4 (Optional[Union[int, float]], optional): k4 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
            k5 (Optional[Union[int, float]], optional): k5 radial distortion coefficient (Brown-Conrady). Defaults to 0.
            k6 (Optional[Union[int, float]], optional): k6 radial distortion coefficient (Brown-Conrady). Defaults to 0.
            p1 (Optional[Union[int, float]], optional): p1 tangential distortion coefficient (Brown-Conrady). Defaults to 0.
            p2 (Optional[Union[int, float]], optional): p2 tangential distortion coefficient (Brown-Conrady). Defaults to 0.
            s1 (Optional[Union[int, float]], optional): s1 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
            s2 (Optional[Union[int, float]], optional): s2 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
            s3 (Optional[Union[int, float]], optional): s3 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
            s4 (Optional[Union[int, float]], optional): s4 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
            skew (Optional[Union[int, float]], optional): camera skew coefficient (fisheye). Defaults to 0.
            parent_frame_id (Optional[str], optional): String id of the parent coordinate frame. Defaults to None.
        """

        if not isinstance(coord_frame_id, str):
            raise Exception("coord_frame_id must be a string")

        if coord_frame_id == "world":
            raise Exception("coord_frame_id cannot be world")

        if self._coord_frame_exists(coord_frame_id):
            raise Exception("Coordinate frame already exists.")

        # If world doesn't exist, make the world coordinate frame
        if not self._coord_frame_exists("world"):
            self._add_coordinate_frame(
                {
                    "coordinate_frame_id": "world",
                    "coordinate_frame_type": "WORLD",
                }
            )

        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        if type(position) is not dict:
            raise Exception("position improperly formatted")
        for position_key in ALL_POSITION_KEYS:
            if not is_valid_number(position[position_key]):
                raise Exception("position coordinates must be valid numbers")

        if orientation is None:
            orientation = {"w": 1, "x": 0, "y": 0, "z": 0}
        if type(orientation) is not dict:
            raise Exception("orientation improperly formatted")
        for orientation_key in ALL_ORIENTATION_KEYS:
            if not is_valid_number(orientation[orientation_key]):
                raise Exception("orientation coordinates must be valid numbers")

        if camera_matrix is not None:
            if type(camera_matrix) != list:
                raise Exception("camera matrix must be a python list of lists")
            if len(camera_matrix) != 4:
                raise Exception("camera matrix is not a properly formatted 4x4 matrix")
            for row in camera_matrix:
                if type(camera_matrix) != list:
                    raise Exception("camera matrix must be a python list of lists")
                if len(row) != 4:
                    raise Exception(
                        "camera matrix is not a properly formatted 4x4 matrix"
                    )
                for el in row:
                    if not is_valid_number(el):
                        raise Exception(
                            "element within camera matrix not a valid float/int"
                        )

        if parent_frame_id is None:
            parent_frame_id = "world"

        if not isinstance(parent_frame_id, str):
            raise Exception("parent_frame_id must be a string or None.")

        if not is_valid_number(fx) or not is_valid_number(fy):
            raise Exception("focal lengths are required and must be valid numbers")

        if fx == 0 or fy == 0:
            raise Exception("focal lengths cannot be 0")

        if camera_model is None:
            camera_model = "brown_conrady"

        if not isinstance(camera_model, str) or camera_model not in ALL_CAMERA_MODELS:
            raise Exception(
                "invalid camera model, valid values are {}".format(ALL_CAMERA_MODELS)
            )

        for coord in [k1, k2, k3, k4, k5, k6, p1, p2, cx, cy, s1, s2, s3, s4, skew]:
            if coord is not None and not is_valid_number(coord):
                raise Exception(
                    f"{coord} is not a valid 2D coordinate frame attribute. Must be float/int/None"
                )

        if (cx is None and cy is not None) or (cx is not None and cy is None):
            raise Exception("optical centers must both be None or both be numbers")
        if (cx is not None and cx <= 0) or (cy is not None and cy <= 0):
            raise Exception("optical centers must be greater than 0")

        metadata: Dict[str, Any] = {
            "fx": fx,
            "fy": fy,
            "camera_model": camera_model,
            "position": position,
            "orientation": orientation,
            "parent_frame_id": parent_frame_id,
        }

        if camera_matrix is not None:
            metadata["camera_matrix"] = camera_matrix
        if k1 is not None:
            metadata["k1"] = k1
        if k2 is not None:
            metadata["k2"] = k2
        if k3 is not None:
            metadata["k3"] = k3
        if k4 is not None:
            metadata["k4"] = k4
        if k5 is not None:
            metadata["k5"] = k5
        if k6 is not None:
            metadata["k6"] = k6
        if p1 is not None:
            metadata["p1"] = p1
        if p2 is not None:
            metadata["p2"] = p2
        if cx is not None:
            metadata["cx"] = cx
        if cy is not None:
            metadata["cy"] = cy
        if s1 is not None:
            metadata["s1"] = s1
        if s2 is not None:
            metadata["s2"] = s2
        if s3 is not None:
            metadata["s3"] = s3
        if s4 is not None:
            metadata["s4"] = s4
        if skew is not None:
            metadata["skew"] = skew

        self._add_coordinate_frame(
            {
                "coordinate_frame_id": coord_frame_id,
                "coordinate_frame_type": "IMAGE_PROJECTION",
                "coordinate_frame_metadata": json.dumps(metadata),
            }
        )

    def add_text(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        text: str,
        date_captured: Optional[str] = None,
    ) -> None:
        """Add a text "sensor" data to this frame.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint (optional if there is only 1 datapoint per frame)
            text (str): The text body.
            date_captured (str, optional): ISO formatted date. Defaults to None.
        """

        if not isinstance(sensor_id, str):
            raise Exception("sensor ids must be strings")

        if date_captured is None:
            date_captured = str(datetime.datetime.now())

        data_urls = {"text": text}
        self._add_coordinate_frame(
            {"coordinate_frame_id": sensor_id, "coordinate_frame_type": "TEXT"}
        )
        self.sensor_data.append(
            {
                "coordinate_frame": sensor_id,
                "data_urls": data_urls,
                "date_captured": date_captured,
                "sensor_id": sensor_id,
                "sensor_metadata": {},
                "sensor_type": "TEXT",
            }
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert this frame into a dictionary representation.

        Returns:
            dict: dictified frame
        """
        row = {
            "task_id": self.frame_id,
            "date_captured": self.date_captured,
            "device_id": self.device_id,
            "coordinate_frames": self.coordinate_frames,
            "sensor_data": self.sensor_data,
            "geo_data": self.geo_data,
            "reuse_latest_embedding": self.reuse_latest_embedding,
        }
        user_metadata_types = {}
        user_metadata_modes = {}

        for k, v, vt, vm in self.user_metadata:
            namespaced = k
            if "user__" not in namespaced:
                namespaced = "user__" + namespaced
            # cast to BQ-serializable list if tuple
            row[namespaced] = list(v) if isinstance(v, tuple) else v
            user_metadata_types[namespaced] = vt
            user_metadata_modes[namespaced] = vm

        row["user_metadata_types"] = user_metadata_types
        row["user_metadata_modes"] = user_metadata_modes
        row["change"] = self.update_type
        return row


class UpdateFrame(BaseFrame):
    is_snapshot: bool

    def __init__(
        self,
        *,
        frame_id: str,
        update_type: UpdateType,
        is_snapshot: bool = True,
        date_captured: Optional[str] = None,
        device_id: Optional[str] = None,
    ):
        super().__init__(
            frame_id=frame_id, date_captured=date_captured, device_id=device_id
        )
        self.is_snapshot = is_snapshot

    def to_dict(self) -> Dict[str, Any]:
        """Convert this frame into a dictionary representation.

        Returns:
            dict: dictified frame
        """

        if self.is_snapshot:
            return super().to_dict()
        else:
            raise Exception("Not Implemented")



"""class_map.py
============
The Labeled Class Map and Class Map Entry Modules and corresponding Update Modules.
"""

from math import floor
from .viridis import viridis_rgb
from .turbo import turbo_rgb
from typing import Any, Optional, List, Dict, Tuple, Union, Set


# https://public.tableau.com/profile/chris.gerrard#!/vizhome/TableauColors/ColorPaletteswithRGBValues
tableau_colors = [
    (31, 119, 180),
    (255, 127, 14),
    (44, 160, 44),
    (214, 39, 40),
    (148, 103, 189),
    (140, 86, 75),
    (227, 119, 194),
    (127, 127, 127),
    (188, 189, 34),
    (23, 190, 207),
]

# https://sashat.me/2017/01/11/list-of-20-simple-distinct-colors/
orig_label_color_list = [
    (230, 25, 75),
    (60, 180, 75),
    (255, 225, 25),
    (67, 99, 216),
    (245, 130, 49),
    (145, 30, 180),
    (70, 240, 240),
    (240, 50, 230),
    (188, 246, 12),
    (250, 190, 190),
    (0, 128, 128),
    (230, 190, 255),
    (154, 99, 36),
    (255, 250, 200),
    (128, 0, 0),
    (170, 255, 195),
    (128, 128, 0),
    (255, 216, 177),
    (0, 0, 117),
    (128, 128, 128),
]

# Used as the "step" between consecutive class colors if the number of classes exceeds the number of scheme colors
DEFAULT_COLOR_STEP = 4

UserProvidedColor = Union[Tuple[int, int, int], List[int]]


class ClassMapEntry:
    """A description of how to interpret a classification.

    In the common case, only three values are needed:
    - The string representation of the class
    - The int [0,255] representation of the class
    - What color to render it as.

    In more complex cases, some classes may be ignored in evaluation,
    or collapsed together into a different class at inference time,
    or tracked as part of a larger category.

    Args:
        name (str): The string representation of the class
        class_id (int): The int representation of the class
        color (Union[Tuple[int, int, int], List[int]]): A length 3 list/tuple of RGB int values
        train_name (Optional[int], optional): The string representation of the class that this label will be inferred as. Defaults to None.
        train_id (Optional[str], optional): The int representation of the class that this label will be infererred as. Defaults to None.
        category (Optional[str], optional): The string representation of the parent category. Defaults to None.
        category_id (Optional[int], optional): The int representation of the parent category. Defaults to None.
        has_instances (bool, optional): Whether each label of this class is a separate instance. Defaults to True.
        ignore_in_eval (bool, optional): Whether to ignore this class while evaluating metrics. Defaults to False.
        train_color (Optional[Union[Tuple[int, int, int], List[int]]], optional): A length 3 list/tuple of RGB int values for showing inferences of this class. Defaults to None.
        keypoint_names (Optional[List[str]], optional): List of keypoint names that exist for this class. Indicies of keypoints in this list are used for the skeletons.
        skeleton (Optional[List[Union[Tuple[int, int], List[int]]]], optional): List of integer pairs that reference indicies in keypoint_names that represent the skeleton for this class.
    """

    def __init__(
        self,
        *,
        name: str,
        class_id: int,
        color: UserProvidedColor,
        train_name: Optional[str] = None,
        train_id: Optional[int] = None,
        category: Optional[str] = None,
        category_id: Optional[int] = None,
        has_instances: bool = True,
        ignore_in_eval: bool = False,
        train_color: Optional[UserProvidedColor] = None,
        keypoint_names: Optional[List[str]] = None,
        skeleton: Optional[List[Union[Tuple[int, int], List[int]]]] = None,
    ) -> None:

        # Set defaults for more complex class mapping fields if not set
        if train_name is None:
            train_name = name
        if train_id is None:
            train_id = class_id
        if category is None:
            category = name
        if category_id is None:
            category_id = class_id
        if train_color is None:
            train_color = color

        # Assert on types
        # Names
        if type(name) is not str:
            raise Exception("Argument 'name' must be a string.")
        if type(train_name) is not str:
            raise Exception("Argument 'train_name' must be a string.")
        if type(category) is not str:
            raise Exception("Argument 'category' must be a string.")

        # Class IDs
        if type(class_id) is not int:
            raise Exception("Argument 'class_id' must be an int.")
        if type(train_id) is not int:
            raise Exception("Argument 'train_id' must be an int.")
        if type(category_id) is not int:
            raise Exception("Argument 'category_id' must be an int.")
        if class_id < 0:
            raise Exception("Argument 'class_id' cannot have negative values.")
        if train_id < 0:
            raise Exception("Argument 'train_id' cannot have negative values.")
        if category_id < 0:
            raise Exception("Argument 'category_id' cannot have negative values.")

        # Colors
        if type(color) not in [list, tuple]:
            raise Exception(
                "Argument 'color' must be a list or tuple of ints, in the range [0,255]."
            )
        if type(train_color) not in [list, tuple]:
            raise Exception(
                "Argument 'train_color' must be a list or tuple of ints, in the range [0,255]."
            )
        if len(color) != 3:
            raise Exception("Argument 'color' must have a length of 3.")
        if len(train_color) != 3:
            raise Exception("Argument 'train_color' must have a length of 3.")
        for val in color:
            if (type(val) is not int) or val < 0 or val > 255:
                raise Exception(
                    "Argument 'color' must have integer entries in the range [0,255]."
                )
        for val in train_color:
            if (type(val) is not int) or val < 0 or val > 255:
                raise Exception(
                    "Argument 'train_color' must have integer entries in the range [0,255]."
                )

        # Keypoints skeleton
        if keypoint_names is not None:
            if not isinstance(keypoint_names, list):
                raise Exception("Argument 'keypoint_names' must be a list of strings.")
            for keypoint_name in keypoint_names:
                if not isinstance(keypoint_name, str):
                    raise Exception(
                        "Argument 'keypoint_names' must be a list of strings."
                    )
        if skeleton is not None:
            if keypoint_names is None:
                raise Exception(
                    "Argument 'keypoint_names' cannot be None when using Argument 'skeleton'."
                )
            if not isinstance(skeleton, list):
                raise Exception("Argument 'skeleton' must be a list of integer pairs.")
            for bone in skeleton:
                if type(bone) not in [list, tuple] or len(bone) != 2:
                    raise Exception(
                        "Argument 'skeleton' must be a list of integer pairs."
                    )
                for i in bone:
                    if i >= len(keypoint_names):
                        raise Exception("Index out of range in argument 'skeleton'.")

        self.name = name
        self.class_id = class_id
        self.train_name = train_name
        self.train_id = train_id
        self.category = category
        self.category_id = category_id
        self.has_instances = has_instances
        self.ignore_in_eval = ignore_in_eval
        self.color = list(color)
        self.train_color = list(train_color)
        self.keypoint_names = keypoint_names
        self.skeleton = skeleton

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictified form of this data structure.

        Returns:
            Dict[str, Any]: dict represenation of Class Map Entry
        """
        return {
            "name": self.name,
            "id": self.class_id,
            "train_name": self.train_name,
            "train_id": self.train_id,
            "category": self.category,
            "category_id": self.category_id,
            "has_instances": self.has_instances,
            "ignore_in_eval": self.ignore_in_eval,
            "color": self.color,
            "train_color": self.train_color,
            "keypoint_names": self.keypoint_names,
            "skeleton": self.skeleton,
        }


class LabelClassMap:
    """A collection of ClassMapEntries that defines how to interpret classifications.

    Args:
        entries (list of ClassMapEntry, optional): List of classification entries. Defaults to None.

    """

    @staticmethod
    def from_classnames(classnames: List[str]) -> "LabelClassMap":
        """A helper utility to generate a set of distinct colors given a list of classnames based on
        the number of classes

        Args:
            classnames (List[str]): List of classifications as strings

        Returns:
            LabelClassMap: A LabelClassMap containing categorical colors for the provided classnames.
        """
        if len(classnames) <= 10:
            return LabelClassMap.from_classnames_max10(classnames)
        elif len(classnames) <= 20:
            return LabelClassMap.from_classnames_max20(classnames)
        else:
            return LabelClassMap.from_classnames_turbo(classnames)

    @staticmethod
    def from_classnames_max20(classnames: List[str]) -> "LabelClassMap":
        """A helper utility to generate a set of distinct colors given a list of classnames

        Args:
            classnames (List[str]): List of up to twenty classifications as strings

        Returns:
            LabelClassMap: A LabelClassMap containing categorical colors for the provided classnames.
        """
        if len(classnames) > 20:
            raise Exception("More than 20 classnames were provided.")

        classmap = LabelClassMap()
        for idx, classname in enumerate(classnames):
            entry = ClassMapEntry(
                class_id=idx,
                name=classname,
                color=orig_label_color_list[idx % len(orig_label_color_list)],
            )
            classmap.add_entry(entry)

        return classmap

    @staticmethod
    def from_classnames_max10(classnames: List[str]) -> "LabelClassMap":
        """A helper utility to generate a set of distinct colors given a list of classnames

        Args:
            classnames (list of str): List of up to ten classifications as strings

        Returns:
            LabelClassMap: A LabelClassMap containing categorical colors for the provided classnames.
        """
        if len(classnames) > 10:
            raise Exception("More than 10 classnames were provided.")

        classmap = LabelClassMap()
        for idx, classname in enumerate(classnames):
            entry = ClassMapEntry(
                class_id=idx,
                name=classname,
                color=tableau_colors[idx % len(tableau_colors)],
            )
            classmap.add_entry(entry)

        return classmap

    @staticmethod
    def from_classnames_turbo(classnames: List[str]) -> "LabelClassMap":
        """A helper utility to generate a set of distinct colors given a list of classnames.

        These colors are pulled from the turbo color scheme, and will be assigned
        in order from dark to light.

        Args:
            classnames (list of str): List of classifications as strings.

        Returns:
            LabelClassMap: A LabelClassMap containing colors for the provided classnames.
        """
        classmap = LabelClassMap()
        col_step = max(DEFAULT_COLOR_STEP, floor(len(turbo_rgb) / len(classnames)))

        for idx, classname in enumerate(classnames):
            col_idx = (col_step * idx) % len(turbo_rgb)
            color = turbo_rgb[col_idx]

            entry = ClassMapEntry(class_id=idx, name=classname, color=color)
            classmap.add_entry(entry)

        return classmap

    @staticmethod
    def from_classnames_viridis(classnames: List[str]) -> "LabelClassMap":
        """A helper utility to generate a set of distinct colors given a list of classnames.

        These colors are pulled from the viridis color scheme, and will be assigned
        in order from dark to light.

        Args:
            classnames (list of str): List of classifications as strings.

        Returns:
            LabelClassMap: A LabelClassMap containing colors for the provided classnames.
        """
        classmap = LabelClassMap()
        col_step = max(DEFAULT_COLOR_STEP, floor(len(viridis_rgb) / len(classnames)))

        for idx, classname in enumerate(classnames):
            col_idx = (col_step * idx) % len(viridis_rgb)
            color = viridis_rgb[col_idx]

            entry = ClassMapEntry(class_id=idx, name=classname, color=color)
            classmap.add_entry(entry)

        return classmap

    def __init__(self, *, entries: Optional[List[ClassMapEntry]] = None) -> None:
        self.entries: List[ClassMapEntry] = []
        self._name_set: Set[str] = set()
        self._class_id_set: Set[int] = set()

        if entries is not None:
            for entry in entries:
                self.add_entry(entry)

    def add_entry(self, entry: ClassMapEntry) -> None:
        """Add a ClassMapEntry.

        Args:
            entry (ClassMapEntry): entry
        """
        if not isinstance(entry, ClassMapEntry):
            raise Exception("entry must be a ClassMapEntry")

        if entry.class_id in self._class_id_set:
            raise Exception(f"An entry with class id {entry.class_id} already exists.")
        if entry.name in self._name_set:
            raise Exception(f"An entry with name {entry.name} already exists.")

        self._class_id_set.add(entry.class_id)
        self._name_set.add(entry.name)

        self.entries.append(entry)


class ClassMapUpdateEntry:
    """A description of a change to be made to a classification. Classification can be referred to by name or id

    Args:
        name (Optional[str], optional): The string representation of the class
        class_id (Optional[int], optional): The int representation of the class
        color (Optional[Union[Tuple[int, int, int], List[int]]], optional): A length 3 list/tuple of RGB int values
        train_color (Optional[Union[Tuple[int, int, int], List[int]]], optional): A length 3 list/tuple of RGB int values for showing inferences of this class. Defaults to None.
        keypoint_names (Optional[List[str]], optional): List of keypoint names that exist for this class. Indicies of keypoints in this list are used for the skeletons.
        skeleton (Optional[List[Union[Tuple[int, int], List[int]]]], optional): List of integer pairs that reference indicies in keypoint_names that represent the skeleton for this class.
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        class_id: Optional[int] = None,
        color: Optional[UserProvidedColor] = None,
        train_color: Optional[UserProvidedColor] = None,
        keypoint_names: Optional[List[str]] = None,
        skeleton: Optional[List[Union[Tuple[int, int], List[int]]]] = None,
    ) -> None:
        # Assert on types
        # Name
        if name:
            if type(name) is not str:
                raise Exception("Argument 'name' must be a string.")

        # Class ID
        if class_id:
            if type(class_id) is not int:
                raise Exception("Argument 'class_id' must be an int.")
            if class_id < 0:
                raise Exception("Argument 'class_id' cannot have negative values.")

        # Colors
        if color:
            if type(color) not in [list, tuple]:
                raise Exception(
                    "Argument 'color' must be a list or tuple of ints, in the range [0,255]."
                )
            if len(color) != 3:
                raise Exception("Argument 'color' must have a length of 3.")
            for val in color:
                if (type(val) is not int) or val < 0 or val > 255:
                    raise Exception(
                        "Argument 'color' must have integer entries in the range [0,255]."
                    )

        if train_color:
            if type(train_color) not in [list, tuple]:
                raise Exception(
                    "Argument 'train_color' must be a list or tuple of ints, in the range [0,255]."
                )

            if len(train_color) != 3:
                raise Exception("Argument 'train_color' must have a length of 3.")

            for val in train_color:
                if (type(val) is not int) or val < 0 or val > 255:
                    raise Exception(
                        "Argument 'train_color' must have integer entries in the range [0,255]."
                    )

        # Some kind of IDness
        if not name and not class_id:
            raise Exception(
                "ClassMapUpdateEntry must have either a 'name' or 'class_id' argument"
            )

        # Some kind of actual change
        if not color and not train_color:
            raise Exception(
                "ClassMapUpdateEntry must include an update of either 'color' or 'train_color'"
            )

        # Keypoints skeleton
        if keypoint_names is not None:
            if not isinstance(keypoint_names, list):
                raise Exception("Argument 'keypoint_names' must be a list of strings.")
            for keypoint_name in keypoint_names:
                if not isinstance(keypoint_name, str):
                    raise Exception(
                        "Argument 'keypoint_names' must be a list of strings."
                    )
        if skeleton is not None:
            if keypoint_names is None:
                raise Exception(
                    "Argument 'keypoint_names' cannot be None when using Argument 'skeleton'."
                )
            if not isinstance(skeleton, list):
                raise Exception("Argument 'skeleton' must be a list of integer pairs.")
            for bone in skeleton:
                if type(bone) not in [list, tuple] or len(bone) != 2:
                    raise Exception(
                        "Argument 'skeleton' must be a list of integer pairs."
                    )
                for i in bone:
                    if i >= len(keypoint_names):
                        raise Exception("Index out of range in argument 'skeleton'.")

        self.name = name
        self.class_id = class_id
        self.color = color
        self.train_color = train_color
        self.keypoint_names = keypoint_names
        self.skeleton = skeleton

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictified form of this data structure.

        Returns:
            Dict[str, Any]: dict represenation of Class Map Update Entry
        """

        return {
            "name": self.name,
            "id": self.class_id,
            "color": self.color,
            "train_color": self.train_color,
            "keypoint_names": self.keypoint_names,
            "skeleton": self.skeleton,
        }



import requests
from typing import Any, Union, Iterable, Dict, List, Callable
from typing_extensions import Literal
from google.cloud import storage  # type: ignore
from http import HTTPStatus

# from .dataset import LabeledDataset

UrlCheckResult = Literal["SKIPPED", "FAILURE", "SUCCESS", "S3_PATH", "INVALID_URL"]


def check_urls(
    urls: Iterable[str],
    gs_client: Any,
    requesting_func: Callable[..., requests.Response] = requests.get,
) -> Dict[str, UrlCheckResult]:
    result = {}
    for url in urls:
        result[url] = check_url(url, gs_client, requesting_func=requesting_func)
    return result


def check_url(
    url: str,
    gs_client: Any,
    requesting_func: Callable[..., requests.Response] = requests.get,
) -> UrlCheckResult:
    if url.lower().startswith("gs://"):
        if gs_client is None:
            return "SKIPPED"
        else:
            bucket_name, path = url.replace("gs://", "").split("/", 1)
            try:
                bucket = gs_client.get_bucket(bucket_name)
                blob = bucket.get_blob(path)

                if blob is None:
                    return "FAILURE"
                else:
                    return "SUCCESS"
            except:
                return "FAILURE"
    elif url.lower().startswith("http"):
        try:
            # we use stream=True to avoid downloading the entire file for SSRF check
            with requesting_func(url, timeout=5, stream=True) as response:
                code = response.status_code
                if code == HTTPStatus.OK:
                    return "SUCCESS"
                else:
                    return "FAILURE"
        except:
            return "FAILURE"
    # We support S3 paths but only the user can access them from the app, and only if they provide their S3 creds
    # See client/src/util/bucketSigning.tsx
    elif url.lower().startswith("s3://"):
        return "S3_PATH"
    else:
        return "INVALID_URL"


UrlMode = Literal["ANONYMOUS", "INACCESSIBLE", "VALID", "INVALID_URL", "S3_PATH"]


def get_mode(
    urls: Iterable[str],
    local_results: Dict[str, UrlCheckResult],
    server_results: Dict[str, UrlCheckResult],
) -> Dict[str, UrlMode]:
    modes: Dict[str, UrlMode] = {}
    for url in urls:
        if server_results[url] == "FAILURE":
            if local_results[url] == "SUCCESS":
                modes[url] = "ANONYMOUS"
            elif local_results[url] == "SKIPPED":
                modes[url] = "INACCESSIBLE"
            elif local_results[url] == "FAILURE":
                modes[url] = "INACCESSIBLE"
        elif server_results[url] == "SUCCESS":
            modes[url] = "VALID"
        elif server_results[url] == "INVALID_URL":
            modes[url] = "INVALID_URL"
        elif server_results[url] == "S3_PATH":
            modes[url] = "S3_PATH"
    return modes


def get_errors(mode: Dict[str, UrlMode], dataset: Any) -> List[str]:
    errors = []

    is_anon = False
    for url in mode:
        if mode[url] == "INVALID_URL":
            errors += [
                f"{url} is not a supported URL format. Please provide a URL starting with gs://, s3://, http://, or https://"
            ]
        elif mode[url] == "INACCESSIBLE":
            errors += [
                f"{url} is not an accessible URL from Aquarium servers or this machine. Please verify the URL path."
            ]
        elif mode[url] == "S3_PATH":
            is_anon = True
            errors += [
                f"{url} is not an accessible URL from Aquarium servers. You will need to set your S3 creds in the web app to view your assets."
            ]
        elif mode[url] == "ANONYMOUS":
            is_anon = True

    if is_anon:
        if (
            len(dataset._temp_frame_embeddings_file_names) == 0
            and len(dataset._temp_frame_embeddings_file_names_streaming) == 0
        ):
            errors += [
                "Anonymous mode detected without embeddings. Please provide your own embeddings. Aquarium cannot compute embeddings in anonymous mode since images are inaccessible from the server."
            ]

    return errors
